#region MyJScript - LGPL Copyright (c) 2006 Lionel Lask
//
// This file is part of MyJScript.
//
// MyJScript is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// MyJScript is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with MyJScript; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
//================================================================================
#endregion

using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.IO;
using System.Text;
using Microsoft.CSharp;
using Microsoft.VisualBasic;

namespace MyJScript.Compiler
{
    /// <summary>
    /// Compiler for MyJScript.
    /// </summary>
    public class Compiler
    {
		/// <summary>
		/// Compile a file.
		/// </summary>
		/// <param name="options">Options to use</param>
		/// <param name="fileName">File to compile</param>
		/// <returns>an object with the compiler results</returns>
        public static CompilerResults CompileAssemblyFromFile(CompilerParameters options, string fileName)
        {
            // Must have parameters
            if (options == null || fileName == null)
            {
                CompilerResults error = new CompilerResults(null);
                error.Errors.Add(new CompilerError(fileName, 1, 1, "MJS0500", "Illegal parameter for compiler"));
                return error;
            }

            // Read file to end
            StreamReader reader = new StreamReader(fileName);
            string source = reader.ReadToEnd();

            // Do compilation
            return DoBuild(fileName, source, options);
        }

		/// <summary>
		/// Compile a source code.
		/// </summary>
		/// <param name="options">Options to use</param>
		/// <param name="source">source code to compile</param>
		/// <returns>an object with the compiler results</returns>
        public static CompilerResults CompileAssemblyFromSource(CompilerParameters options, string source)
        {
            // Must have parameters
            if (options == null || source == null)
            {
                CompilerResults error = new CompilerResults(null);
                error.Errors.Add(new CompilerError(" ", 1, 1, "MJS0500", "Illegal parameter for compiler"));
                return error;
            }

            // Do compilation
            return DoBuild(" ", source, options);
        }

        /// <summary>
        /// Compile source code from a source or a file.
        /// </summary>
        /// <param name="name">name of file</param>
        /// <param name="source">source code</param>
        /// <param name="options">compiler options</param>
        /// <returns>an object with the compiler results</returns>
        private static CompilerResults DoBuild(string name, string source, CompilerParameters options)
        {
			// Create parser and parse
			Parser parser = new Parser(name, source);
			if ( !parser.Parse() )
			{
				CompilerResults error = new CompilerResults(null);
				error.Errors.AddRange(parser.Errors);
				return error;
			}
			CodeNamespace cns = parser.Result;
			
            // Choose between C# and VB compiler
            CodeDomProvider provider;
            if (options.CompilerOptions == "vb")
                provider = new VBCodeProvider();
            else
                provider = new CSharpCodeProvider();

            // If debug, generate C# or VB equivalent
            if (options.IncludeDebugInformation)
            {
                // Create StreamWriter
                Stream s = File.Open(options.OutputAssembly + "." + provider.FileExtension, FileMode.Create);
                StreamWriter sw = new StreamWriter(s);

                // Generate source
                ICodeGenerator generator = provider.CreateGenerator(sw);
                CodeGeneratorOptions cop = new CodeGeneratorOptions();
                generator.GenerateCodeFromNamespace(cns, sw, cop);

                // Close StreamWriter
                sw.Close();
                s.Close();
            }

            // Set options
            CompilerParameters parameters = new CompilerParameters();
            parameters.TreatWarningsAsErrors = false;
            parameters.WarningLevel = 0;
            parameters.ReferencedAssemblies.Add("MJSCRuntime.dll");
            parameters.IncludeDebugInformation = options.IncludeDebugInformation;
            if (options.GenerateInMemory)
            {
                parameters.OutputAssembly = "MemoryScript.dll";
                parameters.GenerateInMemory = true;
                parameters.GenerateExecutable = false;
            }
            else
            {
                parameters.OutputAssembly = options.OutputAssembly;
                parameters.GenerateInMemory = false;
                parameters.GenerateExecutable = true;
                parameters.MainClass = "MyJScript.UserScript.MJSProgram";
            }

            // Use the native compiler to compile CodeDOM
            CodeCompileUnit cu = new CodeCompileUnit();
            cu.Namespaces.Add(cns);
            return provider.CompileAssemblyFromDom(parameters, cu);
        }
    }
}
