#region MyJScript - LGPL Copyright (c) 2006 Lionel Lask
//
// This file is part of MyJScript.
//
// MyJScript is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// MyJScript is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with MyJScript; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
//================================================================================
#endregion

using System;
using System.CodeDom;
using System.Collections.Generic;
using System.Reflection;
using System.Text;

namespace MyJScript.Compiler
{
    /// <summary>
    /// Code generator for MyJScript
    /// </summary>
    public class Generator
    {
        /// <summary>
        /// Name of generated class
        /// </summary>
        private string generatedClassName;

        /// <summary>
        /// Current generated namespace
        /// </summary>
        public CodeNamespace Result
        {
            get { return generatedNamespace; }
        }
        private CodeNamespace generatedNamespace;

        /// <summary>
        /// Current generated class
        /// </summary>
        private CodeTypeDeclaration generatedClass;

        /// <summary>
        /// Main method for class
        /// </summary>
        private CodeMemberMethod mainMethod;

        /// <summary>
        /// Current method
        /// </summary>
        private CodeMemberMethod currentMethod;

        /// <summary>
        /// Declared functions
        /// </summary>
        private Dictionary<String, CodeMemberMethod> functions;

        /// <summary>
        /// Local variables
        /// </summary>
        private Dictionary<String, CodeVariableDeclarationStatement> localVariables;

        /// <summary>
        /// Global variables
        /// </summary>
        private Dictionary<String, CodeMemberField> globalVariables;

        /// <summary>
        /// Statements found
        /// </summary>
        private List<CodeStatement> statements;

        /// <summary>
        /// Block limit in statements list
        /// </summary>
        private CodeStatement blockLimit = new CodeCommentStatement();

        /// <summary>
        /// Function call values
        /// </summary>
        private List<CodeExpression> exprs;

        /// <summary>
        /// Expr limit in expression list
        /// </summary>
        private CodeExpression exprLimit = new CodeSnippetExpression();

        /// <summary>
        /// Object to report error
        /// </summary>
        private Report report;

        /// <summary>
        /// Current location
        /// </summary>
        private Location location;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="className">class name</param>
        /// <param name="reportObject">object to report error</param>
        /// <param name="lexerLocation">location of current token</param>
        public Generator(string className, Report reportObject, Location lexerLocation)
        {
            generatedClassName = className;
            report = reportObject;
            location = lexerLocation;
            currentMethod = null;
            functions = new Dictionary<string, CodeMemberMethod>();
            globalVariables = new Dictionary<string, CodeMemberField>();
            localVariables = new Dictionary<string, CodeVariableDeclarationStatement>();
            statements = new List<CodeStatement>();
            exprs = new List<CodeExpression>();
        }

        /// <summary>
        /// Initialize generator: create a namespace and a class
        /// </summary>
        private void Initialize()
        {
            // Create Namespace
            generatedNamespace = new CodeNamespace("MyJScript.UserScript");

            generatedNamespace.Comments.Add(new CodeCommentStatement(""));
            generatedNamespace.Comments.Add(new CodeCommentStatement(String.Format("  Source generated by MyJScript compiler on {0}", DateTime.Now.ToString())));
            generatedNamespace.Comments.Add(new CodeCommentStatement(""));

            generatedNamespace.Imports.Add(new CodeNamespaceImport("System"));
            generatedNamespace.Imports.Add(new CodeNamespaceImport("MyJScript.Runtime"));

            // Create Class
            generatedClass = new CodeTypeDeclaration();
            generatedClass.IsClass = true;
            generatedClass.Name = generatedClassName;
            generatedClass.TypeAttributes = TypeAttributes.Public;

            generatedNamespace.Types.Add(generatedClass);

            // Create Main
            mainMethod = new CodeMemberMethod();
            mainMethod.Name = "Main";
            mainMethod.Attributes = MemberAttributes.Public | MemberAttributes.Static;

            generatedClass.Members.Add(mainMethod);
        }

        /// <summary>
        /// Start a MyJScript program
        /// </summary>
        public void BeginProgram()
        {
            // Create Namespace and Class
            Initialize();
        }

        /// <summary>
        /// End a MyJScript program
        /// </summary>
        public void EndProgram()
        {
            // Declare global variables
            foreach (KeyValuePair<String, CodeMemberField> pair in globalVariables)
                generatedClass.Members.Insert(0, pair.Value);
            globalVariables.Clear();
        }

        /// <summary>
        /// Start a function declaration
        /// </summary>
        /// <param name="name">function name</param>
        /// <param name="arguments">parameters name</param>
        public void BeginFunction(String name, List<String> arguments)
        {
            // This function already exist
            if (functions.ContainsKey(name))
            {
                report.Error(location, Report.ProcedureAlreadyDefinedError, String.Format("{0} is already defined", name));
                return;
            }

            // Create a new method
            currentMethod = new CodeMemberMethod();
            currentMethod.Name = name;
            currentMethod.Attributes = MemberAttributes.Public | MemberAttributes.Static;
            currentMethod.ReturnType = new CodeTypeReference("MJSValue");
            if (arguments.Count > 0)
            {
                // Check that each parameter name is unique
                bool uniqueParameter = true;
                int length = arguments.Count;
                for (int i = 0; i < length; i++)
                    for (int j = 0; j < length; j++)
                        if (i != j && arguments[i].Equals(arguments[j]))
                            uniqueParameter = false;
                if (!uniqueParameter)
                    report.Error(location, Report.ParameterNameMustBeUniqueError, String.Format("Parameter name must be unique for function \"{0}\"",name));

                // Add parameters
                foreach (string parameter in arguments)
                    currentMethod.Parameters.Add(new CodeParameterDeclarationExpression("MJSValue", parameter));
            }

            // Add to functions dictionnary
            functions.Add(name, currentMethod);
        }

        /// <summary>
        /// End a function declaration
        /// </summary>
        public void EndFunction()
        {
            // Add all statements
            CodeStatementCollection blockStatements = EndBlock();
            currentMethod.Statements.AddRange(blockStatements);

            // Add "return null;" statement at end of function if needed
            int count = currentMethod.Statements.Count;
            if (count == 0 || currentMethod.Statements[count-1].GetType() != typeof(CodeMethodReturnStatement))
                currentMethod.Statements.Add(new CodeMethodReturnStatement(new CodePrimitiveExpression()));

            // Declare local variables
            foreach (KeyValuePair<String, CodeVariableDeclarationStatement> pair in localVariables)
                currentMethod.Statements.Insert(0, pair.Value);
            localVariables.Clear();

            // Add method to class
            generatedClass.Members.Add(currentMethod);

            // Not in a method
            currentMethod = null;
        }

        /// <summary>
        /// Start a new block in program
        /// </summary>
        public void BeginBlock()
        {
            statements.Add(blockLimit);
        }

        /// <summary>
        /// End a block
        /// </summary>
        /// <returns>all statements for the block</returns>
        public CodeStatementCollection EndBlock()
        {
            // Look for the begin of block
            int count = statements.Count;
            int first;
            for (first = count - 1; first >= 0; first--)
                if (statements[first] == blockLimit)
                    break;

            // Create an array with all statements
            CodeStatementCollection blockStatements = new CodeStatementCollection();
            for (int i = first + 1; i < count; i++)
                blockStatements.Add(statements[i]);

            // Remove all statements of the block including block limit
            for (int i = first; i < count; i++)
                statements.RemoveAt(statements.Count - 1);

            return blockStatements;
        }

        /// <summary>
        /// Add statement to current block or to main depending on context.
        /// </summary>
        /// <param name="statement">statement to add</param>
        public void AddStatement(CodeStatement statement)
        {
            // If global scope and not in a block, add to main
            if (currentMethod == null && statements.Count == 0)
                mainMethod.Statements.Add(statement);

            // Add to current block
            else
                statements.Add(statement);
        }

        /// <summary>
        /// Start a list of expression.
        /// </summary>
        public void BeginListExpr()
        {
            exprs.Add(exprLimit);
        }

        /// <summary>
        /// End a list of expression.
        /// </summary>
        /// <returns>the list for this level</returns>
        public List<CodeExpression> EndListExpr()
        {
            // Look for the begin of list
            int count = exprs.Count;
            int first;
            for (first = count - 1; first >= 0; first--)
                if (exprs[first] == exprLimit)
                    break;

            // Create a list with all expressions
            List<CodeExpression> currentList = new List<CodeExpression>();
            for (int i = first + 1; i < count; i++)
                currentList.Add(exprs[i]);

            // Remove all statements of the block including block limit
            for (int i = first; i < count; i++)
                exprs.RemoveAt(exprs.Count - 1);

            return currentList;
        }

        /// <summary>
        /// Add expression in current list.
        /// </summary>
        /// <param name="expr">expression to add</param>
        public void AddExpression(CodeExpression expr)
        {
            exprs.Add(expr);
        }

        /// <summary>
        /// Declare a new variable (local or global depending on current context).
        /// </summary>
        /// <param name="name">variable name</param>
        /// <param name="initValue">init value</param>
        /// <returns>the assign statement</returns>
        public CodeStatement DeclareVariable(string name, CodeExpression initValue)
        {
            if (currentMethod != null)
                return DeclareLocalVar(name, initValue);
            else
                return DeclareGlobalVar(name, initValue);
        }

        /// <summary>
        /// Declare a new global variable.
        /// </summary>
        /// <param name="name">variable name</param>
        /// <param name="initValue">init value</param>
        /// <returns>Statement to set the initial value</returns>
        public CodeStatement DeclareGlobalVar(string name, CodeExpression initValue)
        {
            // Ignore if already exist, generate set value
            if (globalVariables.ContainsKey(name))
                return AssignVariable(name, initValue);

            // Create a new field
            CodeMemberField field = new CodeMemberField();
            field.Name = name;
            field.InitExpression =  new CodeObjectCreateExpression("MJSValue", new CodeExpression[] { });
            field.Attributes = MemberAttributes.Public | MemberAttributes.Static;
            field.Type = new CodeTypeReference("MJSValue");

            // Add to current declaration
            globalVariables.Add(name, field);
            
            // Generate set initial value statement
            return AssignVariable(name, initValue);
        }

        /// <summary>
        /// Declare a new local variable.
        /// </summary>
        /// <param name="name">variable name</param>
        /// <param name="initValue">init value</param>
        /// <returns>Statement to set the initial value</returns>
        public CodeStatement DeclareLocalVar(string name, CodeExpression initValue)
        {
            // Ignore if already exist, generate set value
            if (localVariables.ContainsKey(name))
                return AssignVariable(name, initValue);

            // Ignore if a parameter has same name
            foreach (CodeParameterDeclarationExpression param in currentMethod.Parameters)
                if (param.Name == name)
                {
                    // Generate set value if value is not null
                    if (initValue != null)
                        return AssignVariable(name, initValue);
                    else
                        return new CodeSnippetStatement();
                }

            // Create a new field
            CodeVariableDeclarationStatement var = new CodeVariableDeclarationStatement();
            var.Name = name;
            var.Type = new CodeTypeReference("MJSValue");

            // Add to current declaration
            localVariables.Add(name, var);

            // Generate set initial value statement
            return AssignVariable(name, initValue);
        }

        /// <summary>
        /// Get a variable reference.
        /// </summary>
        /// <param name="name">variable name</param>
        /// <returns>reference expression</returns>
        public CodeExpression Variable(string name)
        {
            // Look first in local variables
            if (localVariables.ContainsKey(name))
                return new CodeVariableReferenceExpression(name);

            // Then in current method parameters
            if (currentMethod != null)
            {
                foreach (CodeParameterDeclarationExpression param in currentMethod.Parameters)
                    if (param.Name == name)
                        return new CodeVariableReferenceExpression(name);
            }
            
            // Finally, create a new variable if it don't exist at global scope
            if (!globalVariables.ContainsKey(name))
                DeclareGlobalVar(name, null);

            // Return reference to global variable
            return new CodeFieldReferenceExpression(new CodeTypeReferenceExpression("MJSProgram"), name);
        }

        /// <summary>
        /// Create a string constant from a value.
        /// </summary>
        /// <param name="value">value</param>
        /// <returns>expression</returns>
        public CodeExpression StringConstant(String value)
        {
            return new CodeObjectCreateExpression("MJSValue", new CodeExpression[] { new CodePrimitiveExpression(value) });
        }

        /// <summary>
        /// Create an int constant from a value.
        /// </summary>
        /// <param name="value">value</param>
        /// <returns>expression</returns>
        public CodeExpression NumberConstant(int value)
        {
            return new CodeObjectCreateExpression("MJSValue", new CodeExpression[] { new CodePrimitiveExpression(value) });
        }

        /// <summary>
        /// Create a statement to assign a value to a variable.
        /// </summary>
        /// <param name="name">name of variable</param>
        /// <param name="value">value to assign</param>
        /// <returns>the statement</returns>
        public CodeStatement AssignVariable(String name, CodeExpression value)
        {
            // Assign to the value or to a null object 
            CodeAssignStatement assign;
            if (value == null)
                assign = new CodeAssignStatement(Variable(name), new CodeObjectCreateExpression("MJSValue", new CodeExpression[] { }));
            else
                assign = new CodeAssignStatement(Variable(name), value);

            return assign;
        }

        /// <summary>
        /// Create a function call expression with its parameters.
        /// </summary>
        /// <param name="name">name of function to call</param>
        /// <param name="values">values for function arguments</param>
        /// <returns>the call expression</returns>
        public CodeExpression FunctionCall(String name, List<CodeExpression> values)
        {
            // Special case of "write": call Console.WriteLine
            CodeExpression[] parameters;
            CodeExpression methodCall;
            if (name == "write")
            {
                // Build param list: take only the first argument and call ToString() on it
                parameters = new CodeExpression[1];
                parameters[0] = new CodeMethodInvokeExpression(values[0], "ToString", new CodeExpression[] { });

                // Call Console.WriteLine
                methodCall = new CodeMethodInvokeExpression(
                    new CodeTypeReferenceExpression("Console"),
                    "WriteLine",
                    parameters);
            }

            // It's not "write": call the method in class
            else
            {
                // Check if method exist
                if (!functions.ContainsKey(name))
                {
                    report.Error(location, Report.ProcedureDoesNotExistError, String.Format("Function \"{0}\" unknown", name));
                    return new CodeSnippetExpression();
                }
                CodeMemberMethod method = functions[name];

                // Compute the right number of argument (truncate or add if needed)
                int paramCount = method.Parameters.Count;
                int valuesCount = values.Count;
                parameters = new CodeExpression[paramCount];
                for (int i = 0; i < paramCount; i++)
                {
                    if (i >= valuesCount)
                        parameters[i] = new CodeObjectCreateExpression("MJSValue", new CodeExpression[] { });
                    else
                        parameters[i] = values[i];
                }

                // Call method in current class
                methodCall = new CodeMethodInvokeExpression(
                    new CodeTypeReferenceExpression("MJSProgram"),
                    name,
                    parameters);
            }

           return methodCall;
        }

        /// <summary>
        /// Create a statement from an expression.
        /// </summary>
        /// <param name="value">an expression</param>
        /// <returns>a statement</returns>
        public CodeStatement Statement(CodeExpression value)
        {
            return new CodeExpressionStatement(value);
        }

        /// <summary>
        /// Create a while statement.
        /// </summary>
        /// <param name="condition">end condition</param>
        /// <param name="toRepeat">statements to repeat</param>
        /// <returns>a while statement</returns>
        public CodeStatement While(CodeExpression condition, CodeStatementCollection toRepeat)
        {
            // Create an array for statements
            int count = toRepeat.Count;
            CodeStatement[] loop = new CodeStatement[count];
            toRepeat.CopyTo(loop, 0);

            // Create a new iteration statement
            return new CodeIterationStatement(
                    new CodeSnippetStatement(""),
                    condition,
                    new CodeSnippetStatement(""),
                    loop);
        }

        /// <summary>
        /// Create a return statement with or without a value.
        /// </summary>
        /// <param name="value">value to return</param>
        /// <returns>a return statement</returns>
        public CodeStatement Return(CodeExpression value)
        {
            // Can't return at global scope
            if (currentMethod == null)
            {
                report.Error(location, Report.ReturnAtGlobalScopeError, "You can't use \"return\" at global scope");
                return new CodeSnippetStatement();
            }
            
            // Create return statement
            if (value == null)
                return new CodeMethodReturnStatement(new CodePrimitiveExpression());
            else
                return new CodeMethodReturnStatement(value);
        }

        /// <summary>
        /// Create an if/else statement.
        /// </summary>
        /// <param name="condition">condition</param>
        /// <param name="ifTrue">statements list to execute if condition is true</param>
        /// <param name="ifFalse">statements list to execute if condition is false</param>
        /// <returns>a condition statement</returns>
        public CodeStatement IfElse(CodeExpression condition, CodeStatementCollection ifTrue, CodeStatementCollection ifFalse)
        {
            // Create an array for statements if true
            int count = ifTrue.Count;
            CodeStatement[] ifTrueStatements = new CodeStatement[count];
            ifTrue.CopyTo(ifTrueStatements, 0);

            // No else
            if (ifFalse == null)
                return new CodeConditionStatement(condition, ifTrueStatements);

            // Create an array for statements if false
            count = ifFalse.Count;
            CodeStatement[] ifFalseStatements = new CodeStatement[count];
            ifFalse.CopyTo(ifFalseStatements, 0);

            return new CodeConditionStatement(condition, ifTrueStatements, ifFalseStatements);
        }

        /// <summary>
        /// Create a binary expression from two expression.
        /// </summary>
        /// <param name="first">first expression</param>
        /// <param name="optype">operator</param>
        /// <param name="second">second expression</param>
        /// <returns></returns>
        public CodeExpression BinaryOperator(CodeExpression first, CodeBinaryOperatorType optype, CodeExpression second)
        {
            return new CodeBinaryOperatorExpression(first, optype, second);
        }
    }
}
