#region MyJScript - LGPL Copyright (c) 2006 Lionel Lask
//
// This file is part of MyJScript.
//
// MyJScript is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// MyJScript is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with MyJScript; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
//================================================================================
#endregion

using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;

namespace MyJScript.Compiler
{
    /// <summary>
    /// Lexical analyzer for MyJScript.
    /// </summary>
    public class Lexer : yyParser.yyInput
    {
        /// <summary>
        /// Token types.
        /// </summary>
        public enum TokenType
        {
            /// <summary>
            /// No token left
            /// </summary>
            EOF = 0,

            /// <summary>
            /// Operators
            /// </summary>
            OperatorPlus = 257,
            OperatorMinus,
            OperatorMult,
            OperatorDiv,
            OperatorEqual,
            OperatorEqualEqual,
            OperatorNotEqual,
            OperatorGreater,
            OperatorLess,
            OperatorGreaterOrEqual,
            OperatorLessOrEqual,

            /// <summary>
            /// Separators
            /// </summary>
            LeftPar,
            RightPar,
            LeftBra,
            RightBra,
            Coma,
            SemiColon,

            /// <summary>
            /// Values
            /// </summary>
            Number,
            String,
            Ident,

            /// <summary>
            /// Keywords
            /// </summary>
            Var,
            Function,
            If,
            Else,
            While,
            Return,

            /// <summary>
            /// Unknown token
            /// </summary>
            Error
        }

        /// <summary>
        /// Known keywords
        /// </summary>
        static private Dictionary<string, TokenType> keywords;

        /// <summary>
        /// Patterns for regular expressions
        /// </summary>
        private const string patternNumber = "^[0-9]+";
        private const string patternQuotedString = "^'[^'\r\n]*'";
        private const string patternDoublequotedString = "^\"[^\"\r\n]*\"";
        private const string patternIdent = "^[_A-Za-z][_A-Za-z0-9]*";

        /// <summary>
        /// Source code to analyze.
        /// </summary>
        private string source;

        /// <summary>
        /// Index in source of next token.
        /// </summary>
        private int index;

        /// <summary>
        /// Current Location
        /// </summary>
        public Location Location;

        /// <summary>
        /// Current token value
        /// </summary>
        private object tokenValue;

        /// <summary>
        /// Class initializer.
        /// </summary>
        static Lexer()
        {
            // Initialize known keywords
            keywords = new Dictionary<string, TokenType>();
            keywords.Add("var", TokenType.Var);
            keywords.Add("function", TokenType.Function);
            keywords.Add("if", TokenType.If);
            keywords.Add("else", TokenType.Else);
            keywords.Add("while", TokenType.While);
            keywords.Add("return", TokenType.Return);
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="code">source code</param>
        public Lexer(string code)
        {
            source = code;
            index = 0;
            tokenValue = null;
            Location = new Location(" ", 1, 1);
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="name">file name</param>
        /// <param name="code">source code</param>
        public Lexer(string name, string code)
        {
            source = code;
            index = 0;
            tokenValue = null;
            Location = new Location(name, 1, 1);
        }

        /// <summary>
        /// Strip white space
        /// </summary>
        private void StripWhiteSpace()
        {
            char next = source[index];
            while (next == ' ' || next == '\t' || next == '\r' || next == '\n')
            {
                if (next == '\n')
                {
                    Location.Row++;
                    Location.Col = 1;
                }
                else if (next != '\r')
                    Location.Col++;
                index++;
                if (index >= source.Length)
                    break;
                next = source[index];
            }
        }

        /// <summary>
        /// Test if next token is an operator.
        /// </summary>
        /// <param name="res">operator found</param>
        /// <returns>true if next token is an operator, false else</returns>
        private bool IsOperator(out TokenType res)
        {
            // Get next char
            char next = source[index];
            int nextNextIndex = index+1;
            res = TokenType.Error;

            // Compare to known operator
            tokenValue = next;
            switch (next)
            {
                case '+':
                    res = TokenType.OperatorPlus;
                    break;

                case '-':
                    res = TokenType.OperatorMinus;
                    break;

                case '/':
                    res = TokenType.OperatorDiv;
                    break;

                case '*':
                    res = TokenType.OperatorMult;
                    break;

                case '!':
                    if (nextNextIndex < source.Length && source[nextNextIndex] == '=')
                    {
                        res = TokenType.OperatorNotEqual;
                        tokenValue = "!=";
                        index++; Location.Col++;
                    }
                    break;

                case '=':
                    if (nextNextIndex < source.Length && source[nextNextIndex] == '=')
                    {
                        res = TokenType.OperatorEqualEqual;
                        tokenValue = "==";
                        index++; Location.Col++;
                    }
                    else
                        res = TokenType.OperatorEqual;
                    break;

                case '>':
                    if (nextNextIndex < source.Length && source[nextNextIndex] == '=')
                    {
                        res = TokenType.OperatorGreaterOrEqual;
                        tokenValue = ">=";
                        index++; Location.Col++;
                    }
                    else
                        res = TokenType.OperatorGreater;
                    break;

                case '<':
                    if (nextNextIndex < source.Length && source[nextNextIndex] == '=')
                    {
                        res = TokenType.OperatorLessOrEqual;
                        tokenValue = "<=";
                        index++; Location.Col++;
                    }
                    else
                        res = TokenType.OperatorLess;
                    break;

                default:
                    return false;
            }

            // An operator was found (warning: this test is need, see case of '!' without '=')
            if (res != TokenType.Error)
            {
                index++; Location.Col++;
                return true;
            }

            return false;
        }


        /// <summary>
        /// Test if next token is a punctuation character.
        /// </summary>
        /// <param name="res">symbol found</param>
        /// <returns>true if next token is a punctuation character, false else</returns>
        private bool IsPunct(out TokenType res)
        {
            // Get next char
            char next = source[index];
            int nextNextIndex = index + 1;
            res = TokenType.Error;

            // Compare to known operator
            switch (next)
            {
                case '(':
                    res = TokenType.LeftPar;
                    break;

                case ')':
                    res = TokenType.RightPar;
                    break;

                case '{':
                    res = TokenType.LeftBra;
                    break;

                case '}':
                    res = TokenType.RightBra;
                    break;

                case ';':
                    res = TokenType.SemiColon;
                    break;

                case ',':
                    res = TokenType.Coma;
                    break;

                default:
                    return false;
            }

            // Go to next char
            tokenValue = next;
            index++; Location.Col++;

            return true;
        }
            
        /// <summary>
        /// Test if next token is a number.
        /// </summary>
        /// <returns>true if next token is a number, false else</returns>
        /// <remarks>if true, token value is set to the number value</remarks>
        private bool IsNumber()
        {
            // Test if match with number pattern
            Match match = Regex.Match(source.Substring(index), patternNumber);
            if (!match.Success)
                return false;

            // Convert to int
            int length = match.Length;
            string number = source.Substring(index, length);
            tokenValue = Int32.Parse(number);
            index += length;
            Location.Col += length;

            return true;
        }

        /// <summary>
        /// Test if next token is a string.
        /// </summary>
        /// <returns>true if next token is a string, false else</returns>
        /// <remarks>if true, token value is set to the string value</remarks>
        private bool IsString()
        {
            // Test first if match to pattern with single quote
            string toTest = source.Substring(index);
            Match match = Regex.Match(toTest, patternQuotedString);

            // Then test with pattern with double quote
            if (!match.Success)
            {
                match = Regex.Match(toTest, patternDoublequotedString);
                if (!match.Success)
                    return false;
            }

            // Extract the value without quotes
            int length = match.Length;
            string stringValue = source.Substring(index+1, length-2);
            tokenValue = stringValue;
            index += length;
            Location.Col += length;

            return true;
        }

        /// <summary>
        /// Test if next token is a keyword or an ident.
        /// </summary>
        /// <param name="res">keyword or ident found</param>
        /// <returns>true if next token is a keyword, false else</returns>
        private bool IsIdentOrKeyword(out TokenType res)
        {
            // Test if match with ident pattern
            Match match = Regex.Match(source.Substring(index), patternIdent);
            if (!match.Success)
            {
                res = TokenType.Error;
                return false;
            }

            // Extract the string
            int length = match.Length;
            string name = source.Substring(index, length);
            index += length;
            Location.Col += length;

            // If it's a keyword
            if (keywords.ContainsKey(name))
            {
                res = keywords[name];
                tokenValue = name;
                return true;
            }

            // It's just an ident
            res = TokenType.Ident;
            tokenValue = name;

            return true;
        }

        /// <summary>
        /// Get the next token.
        /// </summary>
        /// <returns>token id</returns>
        public int token()
        {
            // After EOF !
            if (index == -1)
                return (int)TokenType.Error;

            // Check for EOF
            if (!advance())
            {
                index = -1;
                return (int)TokenType.EOF;
            }

            // Find next token
            TokenType nextToken;
            tokenValue = null;

            // Strip white space
            StripWhiteSpace();

            // Check for EOF
            if (!advance())
            {
                index = -1;
                return (int)TokenType.EOF;
            }

            // Look for all sort of token
            if (IsOperator(out nextToken))
                return (int)nextToken;
            else if (IsPunct(out nextToken))
                return (int)nextToken;
            else if (IsNumber())
                return (int)TokenType.Number;
            else if (IsString())
                return (int)TokenType.String;
            else if (IsIdentOrKeyword(out nextToken))
                return (int)nextToken;

            // Unknown
            tokenValue = source[index];
            return (int)TokenType.Error;
        }

        /// <summary>
        /// Test if another token is available.
        /// </summary>
        /// <returns>true if another token is available, false else.</returns>
        public bool advance()
        {
            return source != null && index != -1 && index < source.Length;
        }

        /// <summary>
        /// Get the value of the current token.
        /// </summary>
        /// <returns></returns>
        public Object value()
        {
            return tokenValue;
        }
    }
}
