#region MyJScript - LGPL Copyright (c) 2006 Lionel Lask
//
// This file is part of MyJScript.
//
// MyJScript is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// MyJScript is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with MyJScript; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
//================================================================================
#endregion

using System;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;

namespace MyJScript.Compiler
{
    /// <summary>
    /// Class for MyJScript command line compiler
    /// </summary>
    public class Program
    {
        /// <summary>
        /// Main method
        /// </summary>
        /// <param name="args">command line arguments</param>
        public static void Main(string[] args)
        {
            // Check number of arguments
            string fileName = null;
            bool debug = false;
            bool vb = false;
            string outputFile = null;
            if (args.Length == 0 || args.Length > 3)
            {
                Syntax();
                return;
            }

            // For each argument...
            foreach (string arg in args)
            {
                // If it's an option
                char firstChar = arg[0];
                if (firstChar == '-' || firstChar == '/')
                {
                    // Is it a valid option ?
                    if (arg.Length == 1)
                    {
                        Syntax();
                        return;
                    }

                    // If it's debug, set debug on
                    string withoutFirst = arg.Substring(1);
                    string option = withoutFirst.ToLower();
                    if (option == "debug" || option == "d")
                    {
                        debug = true;
                        continue;
                    }

                    // If vb, set vb
                    if (option == "vb")
                    {
                        vb = true;
                        continue;
                    }

                    // Look for ':'
                    int index = withoutFirst.IndexOf(':');
                    if (index == -1)
                    {
                        Syntax();
                        return;
                    }
                    option = withoutFirst.Substring(0, index);
                    string optionValue = withoutFirst.Substring(index + 1);

                    // Look for output file name
                    if ((option != "output" && option != "o") || optionValue.Length == 0)
                    {
                        Syntax();
                        return;
                    }
                    outputFile = optionValue;
                }

                // else it's the file name
                else
                {
                    // One file name only
                    if (fileName != null)
                    {
                        Syntax();
                        return;
                    }
                    fileName = arg;
                }
            }

            // Check if file exist
            if (!File.Exists(fileName))
            {
                Logo();
                Console.Error.WriteLine("File {0} not found", fileName);
                return;
            }

            // Compute output file if needed
            if (outputFile == null)
            {
                string filePrefix = fileName;
                int length = fileName.Length;
                if (length > 4)
                {
                    string fileext = fileName.Substring(length - 4, 4).ToLower();
                    if (fileext == ".mjs")
                        filePrefix = fileName.Substring(0, length - 4);
                }
                outputFile = filePrefix + ".exe";
            }
            Logo();
            
            // Call compiler
            CompilerParameters options = new CompilerParameters();
            options.OutputAssembly = outputFile;
            options.IncludeDebugInformation = debug;
            options.GenerateExecutable = true;
            if (vb)
                options.CompilerOptions = "vb";
            CompilerResults result = Compiler.CompileAssemblyFromFile(options, fileName);

            // Display error if any
            foreach(CompilerError error in result.Errors)
                Console.Error.WriteLine(error.ToString());
        }

        /// <summary>
        ///  Write MJSC logo
        /// </summary>
        private static void Logo()
        {
            Console.Error.WriteLine("MJSC - MyJScript Compiler for .NET version {0}\r\nLGPL Copyright (c) 2006 Lionel Lask\r\n", Assembly.GetExecutingAssembly().GetName().Version.ToString());
        }

        /// <summary>
        ///  Write MJSC syntax
        /// </summary>
        private static void Syntax()
        {
            Logo();
            Console.Error.WriteLine("MJSC\r\n	[/output:name]       Output file name            (abbreviated form: /o)\r\n	[/debug]             Include debug information   (abbreviated form: /d)\r\n	[/vb]                Use VB compiler instead of C# compiler\r\n	FileName");
        }
    }
}
