#region MyJScript - LGPL Copyright (c) 2006 Lionel Lask
//
// This file is part of MyJScript.
//
// MyJScript is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// MyJScript is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with MyJScript; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
//================================================================================
#endregion

using System;
using System.CodeDom.Compiler;
using System.Text;

namespace MyJScript.Compiler
{
	/// <summary>
	///   This class is used to report errors and warnings t te user.
	/// </summary>
	public class Report
	{
		/// <summary>
		/// Lexical error
		/// </summary>
		public const string LexicalError = "MJS0100";

        /// <summary>
        /// Unexpected end-of-file error
        /// </summary>
        public const string UnexpectedEOFError = "MJS0200";
			
		/// <summary>
		/// Syntax error
		/// </summary>
		public const string SyntaxError = "MJS0300";

        /// <summary>
        /// Parameter must be unique error
        /// </summary>
        public const string ParameterNameMustBeUniqueError = "MJS0400";

        /// <summary>
        /// Procedure already defined error
        /// </summary>
        public const string ProcedureAlreadyDefinedError = "MJS0401";

        /// <summary>
        /// Procedure does not exist error
        /// </summary>
        public const string ProcedureDoesNotExistError = "MJS0402";

        /// <summary>
        /// Procedure return at global scope error
        /// </summary>
        public const string ReturnAtGlobalScopeError = "MJS0403";
		
		/// <summary>
		/// Error encountered
		/// </summary>
		CompilerErrorCollection errors;


		/// <summary>
		/// Constructor
		/// </summary>
		internal Report()
		{
			errors = new CompilerErrorCollection();
		}

		/// <summary>
		/// Get the errors encountered
		/// </summary>
		public CompilerErrorCollection Errors
		{
			get { return errors; }
		}

		/// <summary>
		/// Report an error from the parser.
		/// </summary>
		/// <param name="location">current location</param>
		/// <param name="state">state in grammar</param>
		/// <param name="token">current token</param>
		/// <param name="tokenValue">current token value</param>
		/// <param name="expecting">list of expecting token name</param>
		public void Error(Location location, int state, int token, object tokenValue, string[] expecting)
		{
			// Lexical error
			CompilerError error;
			if (token == Token.ERROR)
				error = new CompilerError(location.Name, location.Row, location.Col, Report.LexicalError, String.Format("Illegal character {0}", tokenValue.ToString()));

			// Unexpected end of file error
			else if (token == Token.EOF)
                error = new CompilerError(location.Name, location.Row, location.Col, Report.UnexpectedEOFError, "Unexpected end of file");

			// Syntax error
			else
			{
				StringBuilder expectings = new StringBuilder();
				int length = expecting.Length;
				for( int i = 0 ; i < length ; i++ )
				{
					expectings.Append(' ');
					expectings.Append(expecting[i]);
				}
                error = new CompilerError(location.Name, location.Row, location.Col, Report.SyntaxError, String.Format("Syntax error on \"{0}\", state {1}, expecting:{2}", tokenValue, state.ToString(), expectings.ToString()));
			}

			errors.Add(error);
		}

		/// <summary>
		/// Report an error.
		/// </summary>
		/// <param name="location">current location</param>
        /// <param name="errorType">error type</param>
		/// <param name="errorText">text to qualify error</param>
		public void Error(Location location, string errorType, string errorText)
		{
            errors.Add(new CompilerError(location.Name, location.Row, location.Col, errorType, errorText));
		}
	}
}

