#region MyJScript - LGPL Copyright (c) 2006 Lionel Lask
//
// This file is part of MyJScript.
//
// MyJScript is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// MyJScript is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with MyJScript; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
//================================================================================
#endregion

using System;
using System.Collections.Generic;
using System.Text;

namespace MyJScript.Runtime
{
    /// <summary>
    /// Class for a MyJScript value.
    /// </summary>
    public class MJSValue : IComparable
    {
        /// <summary>
        /// Value of the object.
        /// </summary>
        public Object Value
        {
            get { return currentValue; }
        }
        private Object currentValue;

        #region Constructors
        /// <summary>
        /// Null constructor.
        /// </summary>
        public MJSValue()
        {
            currentValue = null;
        }

        /// <summary>
        /// Constructor for number.
        /// </summary>
        /// <param name="initValue">initial value</param>
        public MJSValue(float initValue)
        {
            currentValue = initValue;
        }

        /// <summary>
        /// Constructor for string.
        /// </summary>
        /// <param name="initValue">initial value</param>
        public MJSValue(string initValue)
        {
            currentValue = initValue;
        }
        #endregion

        /// <summary>
        /// Test is value is null.
        /// </summary>
        /// <returns>true if value is null, false else</returns>
        public bool IsNull()
        {
            return currentValue == null;
        }

        #region Arithmetic operators
        /// <summary>
        /// Add operator.
        /// </summary>
        /// <param name="first">first parameter</param>
        /// <param name="second">second parameter</param>
        /// <returns>result of first + second or a null value if error</returns>
        public static MJSValue operator +(MJSValue first, MJSValue second)
        {
            if (first.Value == null || second.Value == null)
                return new MJSValue();

            if (first.Value is string || second.Value is string)
                return new MJSValue(first.Value.ToString() + second.Value.ToString());

            if (first.Value is float && second.Value is float)
                return new MJSValue((float)first.Value + (float)second.Value);

            return new MJSValue();
        }

        /// <summary>
        /// Product operator.
        /// </summary>
        /// <param name="first">first parameter</param>
        /// <param name="second">second parameter</param>
        /// <returns>result of first * second or a null value if error</returns>
        public static MJSValue operator *(MJSValue first, MJSValue second)
        {
            // Both are float
            if (first.Value is float && second.Value is float)
                return new MJSValue((float)first.Value * (float)second.Value);

            // If currentValue is null, other value must be null
            if (first.Value == null || second.currentValue == null)
                return new MJSValue();

            // Try to convert to float...
            if (first.Value is float)
            {
                float secondValueFloat;
                if (Single.TryParse(second.currentValue.ToString(), out secondValueFloat))
                    return new MJSValue((float)first.Value * secondValueFloat);
            }
            else if (second.Value is float)
            {
                float firstValueFloat;
                if (Single.TryParse(first.currentValue.ToString(), out firstValueFloat))
                    return new MJSValue(firstValueFloat * (float)second.Value);
            }

            return new MJSValue();
        }

        /// <summary>
        /// Minus operator.
        /// </summary>
        /// <param name="first">first parameter</param>
        /// <param name="second">second parameter</param>
        /// <returns>result of first - second or a null value if error</returns>
        public static MJSValue operator -(MJSValue first, MJSValue second)
        {
            if (first.Value is float && second.Value is float)
                return new MJSValue((float)first.Value - (float)second.Value);

            // If currentValue is null, other value must be null
            if (first.Value == null || second.currentValue == null)
                return new MJSValue();

            // Try to convert to float...
            if (first.Value is float)
            {
                float secondValueFloat;
                if (Single.TryParse(second.currentValue.ToString(), out secondValueFloat))
                    return new MJSValue((float)first.Value - secondValueFloat);
            }
            else if (second.Value is float)
            {
                float firstValueFloat;
                if (Single.TryParse(first.currentValue.ToString(), out firstValueFloat))
                    return new MJSValue(firstValueFloat - (float)second.Value);
            }

            return new MJSValue();
        }

        /// <summary>
        /// Divide operator.
        /// </summary>
        /// <param name="first">first parameter</param>
        /// <param name="second">second parameter</param>
        /// <returns>result of first / second or a null value if error</returns>
        public static MJSValue operator /(MJSValue first, MJSValue second)
        {
            if (first.Value is float && second.Value is float)
                return new MJSValue((float)first.Value / (float)second.Value);

            // If currentValue is null, other value must be null
            if (first.Value == null || second.currentValue == null)
                return new MJSValue();

            // Try to convert to float...
            if (first.Value is float)
            {
                float secondValueFloat;
                if (Single.TryParse(second.currentValue.ToString(), out secondValueFloat))
                    return new MJSValue((float)first.Value / secondValueFloat);
            }
            else if (second.Value is float)
            {
                float firstValueFloat;
                if (Single.TryParse(first.currentValue.ToString(), out firstValueFloat))
                    return new MJSValue(firstValueFloat / (float)second.Value);
            }

            return new MJSValue();
        }
        #endregion

        /// <summary>
        /// Value as string.
        /// </summary>
        /// <returns>value as a string</returns>
        public override string ToString()
        {
            if (currentValue == null)
                return "undefined";
            return currentValue.ToString();
        }

        /// <summary>
        /// Compute hashcode for value.
        /// </summary>
        /// <returns>an integer value reprensenting uniquely the value</returns>
        public override int GetHashCode()
        {
            // TODO: Compute a specific encoding
            return base.GetHashCode();
        }

        #region Comparison operators
        /// <summary>
        /// Test equality.
        /// </summary>
        /// <param name="obj">object to compare</param>
        /// <returns>true if obj is equal to this, false else</returns>
        public override bool Equals(object obj)
        {
            if (obj == null)
                return false;
            MJSValue mjsObj = obj as MJSValue;
            if ((object)mjsObj == null)
                return false;
            return this == mjsObj;
        }

        /// <summary>
        ///  Test equality.
        /// </summary>
        /// <param name="first">first object</param>
        /// <param name="second">second object</param>
        /// <returns>true if first is equal to second, false else</returns>
        public static bool operator ==(MJSValue first, MJSValue second)
        {
            // Null allowed only if both are null
            if ((object)first == null)
                return ((object)second == null);
            if ((object)second == null)
                return false;

            // If one value is null, other must be null
            if (first.currentValue == null)
                return second.currentValue == null;
            if (second.currentValue == null)
                return false;

            // If types are different
            if (first.Value.GetType() != second.Value.GetType())
            {
                // Convert to string
                return first.Value.ToString().Equals(second.Value.ToString());
            }

            // Use comparison for this type
            return first.currentValue.Equals(second.currentValue);
        }

        /// <summary>
        ///  Test inequality.
        /// </summary>
        /// <param name="first">first object</param>
        /// <param name="second">second object</param>
        /// <returns>true if first is not equal to second, false else</returns>
        public static bool operator !=(MJSValue first, MJSValue second)
        {
            return !(first == second);
        }

        /// <summary>
        /// Compare current to object to another.
        /// </summary>
        /// <param name="obj">object to compare</param>
        /// <returns>-1 if current object is lesser than other one, 1 if current object is greater than other one, 0 else</returns>
        public int CompareTo(object obj)
        {
            // Can't compare to null
            if (obj == null)
                throw new ArgumentNullException();

            // Can't compare to other type
            if (!(obj is MJSValue))
                throw new ArgumentException();

            // If currentValue is null, other value must be null
            MJSValue mjsObj = obj as MJSValue;
            if (currentValue == null)
            {
                if (mjsObj.currentValue == null)
                    return 0;
                throw new ArgumentException();
            }
            if (mjsObj.currentValue == null)
                throw new ArgumentException();

            // If types are different
            IComparable value = (IComparable)currentValue;
            if (currentValue.GetType() != mjsObj.currentValue.GetType())
            {
                // Try to convert to float...
                if (currentValue is float)
                {
                    float mjsObjValueFloat;
                    if (Single.TryParse(mjsObj.currentValue.ToString(), out mjsObjValueFloat))
                        return value.CompareTo(mjsObjValueFloat);
                }
                else if (mjsObj.currentValue is float)
                {
                    float valueFloat;
                    if (Single.TryParse(currentValue.ToString(), out valueFloat))
                        return valueFloat.CompareTo(mjsObj.currentValue);
                }

                // ...else, convert both value to string
                return currentValue.ToString().CompareTo(mjsObj.currentValue.ToString());
            }

            // Use comparison for this type
            return value.CompareTo(mjsObj.currentValue);
        }

        /// <summary>
        ///  Greater than.
        /// </summary>
        /// <param name="first">first object</param>
        /// <param name="second">second object</param>
        /// <returns>true if first is greater than second, false else</returns>
        public static bool operator >(MJSValue first, MJSValue second)
        {
            if ((object)first == null || (object)second == null)
                throw new ArgumentNullException();
            return first.CompareTo(second) > 0;
        }

        /// <summary>
        ///  Lesser than.
        /// </summary>
        /// <param name="first">first object</param>
        /// <param name="second">second object</param>
        /// <returns>true if first is lesser than second, false else</returns>
        public static bool operator <(MJSValue first, MJSValue second)
        {
            if ((object)first == null || (object)second == null)
                throw new ArgumentNullException();
            return first.CompareTo(second) < 0;
        }

        /// <summary>
        ///  Greater or equal than.
        /// </summary>
        /// <param name="first">first object</param>
        /// <param name="second">second object</param>
        /// <returns>true if first is greater or equal than second, false else</returns>
        public static bool operator >=(MJSValue first, MJSValue second)
        {
            if ((object)first == null)
            {
                if ((object)second == null)
                    return true;
                throw new ArgumentNullException();
            }
            return first.CompareTo(second) >= 0;
        }

        /// <summary>
        ///  Lesser or equal than.
        /// </summary>
        /// <param name="first">first object</param>
        /// <param name="second">second object</param>
        /// <returns>true if first is lesser or equal than second, false else</returns>
        public static bool operator <=(MJSValue first, MJSValue second)
        {
            if ((object)first == null)
            {
                if ((object)second == null)
                    return true;
                throw new ArgumentNullException();
            }
            return first.CompareTo(second) <= 0;
        }

        #endregion
    }
}
