﻿#region MyJScript - LGPL Copyright (c) 2006 Lionel Laské
//
// This file is part of MyJScript.
//
// MyJScript is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// MyJScript is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with MyJScript; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
//
//================================================================================
#endregion

using System;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using MyJScript.Compiler;

namespace MyJScript.Test
{
    /// <summary>
    /// Unit testing for MyJScript compiler
    /// </summary>
    [TestClass]
    public class CompilerTest
    {
        /// <summary>
        /// Constructor
        /// </summary>
        public CompilerTest()
        {
        }

        /// <summary>
        /// Compile and run a source.
        /// </summary>
        /// <param name="source">source code</param>
        /// <returns>console output</returns>
        private string CompileAndRun(string source)
        {
            // Compile
            CompilerParameters options = new CompilerParameters();
            options.GenerateInMemory = true;
            options.IncludeDebugInformation = false;
            CompilerResults results = Compiler.Compiler.CompileAssemblyFromSource(options, source);
            foreach (CompilerError error in results.Errors)
                Console.Error.WriteLine(error.ToString());
            if (results.Errors.Count > 0)
                return "";

            // Put output stream to a string
            StringWriter outputString = new StringWriter();
            Console.SetOut(outputString);

            // Call the Main method
            Assembly assembly = results.CompiledAssembly;
            Assert.IsNotNull(assembly);
            Type evaluatorType = assembly.GetType("MyJScript.UserScript.MJSProgram");
            Assert.IsNotNull(evaluatorType);
            try
            {
                evaluatorType.InvokeMember("Main", BindingFlags.Public | BindingFlags.Static | BindingFlags.InvokeMethod, null, null, null);
            }
            catch (TargetInvocationException e)
            {
                throw e.InnerException;
            }

            // Reinitialize output stream
            string console = outputString.ToString();
			Stream outputStream = Console.OpenStandardOutput();
            Console.SetOut(new StreamWriter(outputStream));

            return console;
        }

        /// <summary>
        /// Test Empty program.
        /// </summary>
        [TestMethod]
        public void Empty()
        {
            Assert.AreEqual("", CompileAndRun(""));
        }

        /// <summary>
        /// Test write.
        /// </summary>
        [TestMethod]
        public void Write()
        {
            Assert.AreEqual("Hello world!\r\n", CompileAndRun("write('Hello world!');"));
            Assert.AreEqual("100\r\n", CompileAndRun("write(100);"));
            Assert.AreEqual("undefined\r\n", CompileAndRun("write(zzz);"));
            Assert.AreEqual("First\r\n", CompileAndRun("write(\"First\", 'Second', 3);"));
            Assert.AreEqual("Double\"Quote\r\n", CompileAndRun("write('Double\"Quote');"));
        }

        /// <summary>
        /// Test variables.
        /// </summary>
        [TestMethod]
        public void Variables()
        {
            Assert.AreEqual("undefined\r\nLionel\r\n", CompileAndRun("var a; write(a); a = \"Lionel\"; write(a);"));
            Assert.AreEqual("12\r\nGo!\r\n", CompileAndRun("var a = 12; b = 'Go!'; write(a); write(b);"));
            Assert.AreEqual("100\r\n150\r\n200\r\n100\r\n", CompileAndRun("var _x = 100; write(_x); function hide(_x) { write(_x); _x = 200; write(_x); } hide(150); write(_x);"));
            Assert.AreEqual("30\r\n", CompileAndRun("function param(a) { var a; write(a); } param(30);"));
            Assert.AreEqual("69\r\n", CompileAndRun("function param2(a) { var a = 69; write(a); } param2(12);"));
            Assert.AreEqual("XXX\r\n1000\r\n", CompileAndRun("function doubledecl() { var yan = 'XXX'; write(yan); var yan = 1000; write(yan); } doubledecl();"));
        }

        /// <summary>
        /// Test function call.
        /// </summary>
        [TestMethod]
        public void FunctionCall()
        {
            Assert.AreEqual("1\r\nTwo\r\n", CompileAndRun("function tooparam(a,b) { write(a); write(b); } tooparam(1,'Two',3);"));
            Assert.AreEqual("1\r\nundefined\r\n", CompileAndRun("function missparam(a,b) { write(a); write(b); } missparam(1);"));
        }

        /// <summary>
        /// Test arithmetic expression.
        /// </summary>
        [TestMethod]
        public void Arithmetic()
        {
            Assert.AreEqual("2100\r\n", CompileAndRun("write(2+'100');"));
            Assert.AreEqual("Hello world!\r\n", CompileAndRun("var a = 'Hello '; var b = \"world!\"; write(a+b);"));
            Assert.AreEqual("154\r\n", CompileAndRun("var quatre=4; var z = 150; z = z + quatre; write(z);"));

            Assert.AreEqual("55\r\n", CompileAndRun("function foo(sub) { write(100-sub); }foo(45);"));
            Assert.AreEqual("undefined\r\n", CompileAndRun("write('Liogo'-'go');"));
            Assert.AreEqual("-1\r\n", CompileAndRun("var one=1; write(one+one-3);"));

            Assert.AreEqual("16\r\n", CompileAndRun("write(2*2*2*2);"));
            Assert.AreEqual("undefined\r\n", CompileAndRun("var undef = 10*'H'; write(undef);"));

            Assert.AreEqual("5\r\n", CompileAndRun("var six='6'; write(30/six);"));
            Assert.AreEqual("undefined\r\n", CompileAndRun("write('ZZZZ'/4);"));
        }

        /// <summary>
        /// Test while.
        /// </summary>
        [TestMethod]
        public void While()
        {
            Assert.AreEqual("before\r\nafter\r\n", CompileAndRun("write('before'); while (0==1) write('false'); write('after');"));
            Assert.AreEqual("before\r\n0\r\n1\r\n2\r\nafter\r\n", CompileAndRun("var i = 0; write('before'); while (i<3) { write(i); i=i+1; } write('after');"));
            Assert.AreEqual("before\r\n0\r\nA0\r\nA1\r\n1\r\nA0\r\nA1\r\n2\r\nA0\r\nA1\r\nafter\r\n", CompileAndRun("var i = 0; write('before'); while (i<3) { var j = 0; write(i); while(j<2) { write('A'+j); j=j+1; } i=i+1; } write('after');"));

            Assert.AreEqual("before\r\nafter\r\n", CompileAndRun("function foo() { write('before'); while (0==1) write('false'); write('after'); } foo();"));
            Assert.AreEqual("before\r\n0\r\n1\r\n2\r\nafter\r\n", CompileAndRun("function loop(n) { var i = 0; write('before'); while (i<n) { write(i); i=i+1; } write('after'); } loop(3);"));
            Assert.AreEqual("before\r\n0\r\nA0\r\nA1\r\n1\r\nA0\r\nA1\r\n2\r\nA0\r\nA1\r\nafter\r\n", CompileAndRun("function loop() { var i = 0; write('before'); while (i<3) { var j = 0; write(i); while(j<2) { write('A'+j); j=j+1; } i=i+1; } write('after'); } loop();"));
        }

        /// <summary>
        /// Test Return.
        /// </summary>
        [TestMethod]
        public void Return()
        {
            Assert.AreEqual("before\r\n", CompileAndRun("function foo() { write('before'); return; write('after'); } foo();"));
            Assert.AreEqual("before\r\nYes\r\n", CompileAndRun("function foo() { write('before'); return 'Yes'; write('after'); } write(foo());"));
            Assert.AreEqual("before\r\nYes\r\n", CompileAndRun("function foo() { write('before'); while(1==1) { return 'Yes'; } write('after'); } write(foo());"));
            Assert.AreEqual("20\r\n200\r\n", CompileAndRun("function twotime(n) { return 2*n; } write(twotime(10)); write(twotime(100));"));
        }

        /// <summary>
        /// Test IfElse.
        /// </summary>
        [TestMethod]
        public void IfElse()
        {
            Assert.AreEqual("Yes\r\n", CompileAndRun("if(1==1) write('Yes');"));
            Assert.AreEqual("after\r\n", CompileAndRun("if(1==0) write('Yes'); write('after');"));
            Assert.AreEqual("Yes\r\n", CompileAndRun("if(1==1) write('Yes'); else write('No');"));
            Assert.AreEqual("No\r\n", CompileAndRun("if(1==0) write('Yes'); else write('No');"));
            Assert.AreEqual("Yes\r\n2\r\n", CompileAndRun("var i = 1; if(i>0) { write('Yes'); i=i+1; } else { write('No'); i=i-1; } write(i);"));
            Assert.AreEqual("No\r\n-1\r\n", CompileAndRun("var i = 0; if(i>0) { write('Yes'); i=i+1; } else { write('No'); i=i-1; } write(i);"));
            Assert.AreEqual(">10\r\n", CompileAndRun("var i = 34; if(i>10) if (i>100) write('>100'); else write('>10'); else write('<10');"));
            Assert.AreEqual("<10\r\n", CompileAndRun("var i = 8; if(i>10) if (i>100) write('>100'); else write('>10'); else write('<10');"));
        }

        /// <summary>
        /// Test comparison expression.
        /// </summary>
        [TestMethod]
        public void Comparison()
        {
            Assert.AreEqual("No\r\n", CompileAndRun("if(0>1) write('Yes'); else write('No');"));
            Assert.AreEqual("Yes\r\n", CompileAndRun("if(1>0) write('Yes'); else write('No');"));
            Assert.AreEqual("Yes\r\n", CompileAndRun("if(0<1) write('Yes'); else write('No');"));
            Assert.AreEqual("No\r\n", CompileAndRun("if(1<0) write('Yes'); else write('No');"));
            Assert.AreEqual("No\r\n", CompileAndRun("if(0>=1) write('Yes'); else write('No');"));
            Assert.AreEqual("Yes\r\n", CompileAndRun("if(1>=0) write('Yes'); else write('No');"));
            Assert.AreEqual("Yes\r\n", CompileAndRun("if(1>=1) write('Yes'); else write('No');"));
            Assert.AreEqual("Yes\r\n", CompileAndRun("if(0<=1) write('Yes'); else write('No');"));
            Assert.AreEqual("No\r\n", CompileAndRun("if(1<=0) write('Yes'); else write('No');"));
            Assert.AreEqual("Yes\r\n", CompileAndRun("if(1<=1) write('Yes'); else write('No');"));
            Assert.AreEqual("No\r\n", CompileAndRun("if(0==1) write('Yes'); else write('No');"));
            Assert.AreEqual("Yes\r\n", CompileAndRun("if(1==1) write('Yes'); else write('No');"));
            Assert.AreEqual("Yes\r\n", CompileAndRun("if(0!=1) write('Yes'); else write('No');"));
            Assert.AreEqual("No\r\n", CompileAndRun("if(1!=1) write('Yes'); else write('No');"));
        }

        /// <summary>
        /// Test a full program.
        /// </summary>
        [TestMethod]
        public void Program()
        {
            Assert.AreEqual("1\r\n24\r\n120\r\n", CompileAndRun(@"
                function fact(n) {
                    if (n==0) return 1;
                    return n*fact(n-1);
                }

                write(fact(1));
                write(fact(4));
                write(fact(5));
            "));

            Assert.AreEqual("Hello DNG\r\n2\r\n4\r\n6\r\n8\r\n", CompileAndRun(@"
                function twotime(n)
                {
                    return 2*n;
                }

                var i = 1;
                write(""Hello DNG"");
                while(i < 5)
                {
                   write(twotime(i));
                   i = i + 1;
                }
            "));
        }
    }
}
